class RoombaInfo:
    hostname = None
    firmware = None
    ip = None
    mac = None
    robot_name = None
    sku = None
    capabilities = None
    blid = None
    password = None

    def __init__(
        self, hostname, robot_name, ip, mac, firmware, sku, capabilities
    ):
        """Create object with information about roomba."""
        self.hostname = hostname
        self.firmware = firmware
        self.ip = ip
        self.mac = mac
        self.robot_name = robot_name
        self.sku = sku
        self.capabilities = capabilities
        self.blid = hostname.split("-")[1]

    def __str__(self) -> str:
        """Nice output to console."""
        return ", ".join(
            [
                "{key}={value}".format(key=key, value=self.__dict__.get(key))
                for key in self.__dict__
            ]
        )

    def __hash__(self) -> int:
        """Hashcode."""
        return hash(self.mac)

    def __eq__(self, o: object) -> bool:
        """Equals."""
        return isinstance(o, RoombaInfo) and self.mac == o.mac



from typing import Tuple
from .math_helpers import clamp

try:
    from PIL import Image, ImageDraw, ImageFont, ImageColor
    HAVE_PIL = True
except ImportError:
    pass

transparent = (0, 0, 0, 0)  #transparent color

def make_blank_image(width, height, color=transparent) -> Image.Image:
    return Image.new('RGBA',(width,height), color)

def transparent_paste(base_image: Image.Image, overlay_image: Image.Image, position: Tuple = None):
    '''
    needed because PIL pasting of transparent images gives weird results
    '''
    image = make_blank_image(base_image.size[0],base_image.size[1])
    image.paste(overlay_image,position)
    base_image = Image.alpha_composite(base_image, image)
    return base_image

def center_image(ox: int, oy: int, image: Image.Image, bounds: Tuple[int,int]) -> Tuple[int,int]:
    xx, yy = (ox - image.size[0] // 2, oy - image.size[1] // 2)
    if bounds:
        xx = clamp(xx, 0, bounds[0])
        yy = clamp(yy, 0, bounds[1])
        
    return (xx, yy)
            
def validate_color(color, default) -> Tuple[int,int,int,int]:      
    try:
        return ImageColor.getcolor(color,"RGBA")
    except:
        return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Python 3.* (thanks to pschmitt for adding Python 3 compatibility).

Program to connect to Roomba 980 vacuum cleaner, dcode json, and forward to mqtt
server.
Nick Waterton 24th April 2017: V 1.0: Initial Release
Nick Waterton 4th July   2017  V 1.1.1: Fixed MQTT protocol version, and map
paths, fixed paho-mqtt tls changes
Nick Waterton 5th July   2017  V 1.1.2: Minor fixes, CV version 3 .2 support
Nick Waterton 7th July   2017  V1.2.0: Added -o option "roomOutline" allows
enabling/disabling of room outline drawing, added auto creation of css/html
files Nick Waterton 11th July  2017  V1.2.1: Quick (untested) fix for room
outlines if you don't have OpenCV
"""
import asyncio
import datetime
import json
import logging
import threading
import time
from collections import OrderedDict
from collections.abc import Mapping
from datetime import datetime

from roombapy.mapping import (
    DEFAULT_ICON_SIZE,
    RoombaMapper,
    RoombaMap,
    RoombaMapDevice
)

from roombapy.const import (
    ROOMBA_ERROR_MESSAGES, 
    ROOMBA_READY_MESSAGES, 
    ROOMBA_STATES
)
from roombapy.remote_client import RoombaRemoteClient

MAX_CONNECTION_RETRIES = 3


class RoombaConnectionError(Exception):
    """Roomba connection exception."""
    pass

class Roomba:
    """
    This is a Class for Roomba 900 series WiFi connected Vacuum cleaners.

    Requires firmware version 2.0 and above (not V1.0). Tested with Roomba 980
    username (blid) and password are required, and can be found using the
    password() class above (or can be auto discovered)
    Most of the underlying info was obtained from here:
    https://github.com/koalazak/dorita980 many thanks!
    The values received from the Roomba as stored in a dictionay called
    master_state, and can be accessed at any time, the contents are live, and
    will build with time after connection.
    This is not needed if the forward to mqtt option is used, as the events will
    be decoded and published on the designated mqtt client topic.
    """

    def __init__(self, remote_client: RoombaRemoteClient, continuous=True, delay=1):
        """Roomba client initialization."""
        self.log = logging.getLogger(__name__)
        self.loop = asyncio.get_event_loop()        

        self.remote_client = remote_client
        self._init_remote_client_callbacks()
        self.continuous = continuous
        if self.continuous:
            self.log.debug("CONTINUOUS connection")
        else:
            self.log.debug("PERIODIC connection")

        self.stop_connection = False
        self.periodic_connection_running = False
        self.topic = "#"
        self.exclude = ""
        self.delay = delay
        self.periodic_connection_duration = 10
        self.roomba_connected = False
        self.indent = 0
        self.master_indent = 0
        self.current_state = None
        self.master_state = {}  # all info from roomba stored here
        self.time = time.time()
        self._thread = threading.Thread(
            target=self.periodic_connection, name="roombapy"
        )
        self.on_message_callbacks = []
        self.on_disconnect_callbacks = []
        self.client_error = None

        #create the mapper
        self._mapper = RoombaMapper(self)
        self._history = {}
        self._flags = {}
        self._new_mission_start_time: float = None

        self._pmap_id: str = None
        self._maps: dict[str,RoombaMap] = {}        
        self._devices: dict[str,RoombaMapDevice] = {}

    @property    
    def co_ords(self):
        co_ords = self.pose
        if isinstance(co_ords, dict):
            return {'x': co_ords['point']['y'],
                    'y': co_ords['point']['x'],
                    'theta': co_ords['theta']}
        return self.zero_coords()

    @property
    def blid(self):
        return self.remote_client.blid

    @property
    def current_pmap_id(self):
        return self._pmap_id
        
    @property
    def error_num(self):
        try:
            return self.cleanMissionStatus.get('error')
        except AttributeError:
            pass
        return 0
        
    @property
    def error_message(self):
        return self._get_error_message(self.error_num)

    @property
    def not_ready_num(self):
        try:
            return self.cleanMissionStatus.get('notReady')
        except:
            pass
        return 0
    
    @property
    def not_ready_message(self):
        return self._get_not_ready_message(self.not_ready_num)    
        
    @property
    def cleanMissionStatus(self):
        return self.get_property("cleanMissionStatus")
        
    @property
    def pose(self):
        return self.get_property("pose")
        
    @property
    def batPct(self):
        return self.get_property("batPct")
                 
    @property
    def bin_full(self):
        return self.get_property("bin_full")
        
    @property
    def tanklvl(self):
        return self.get_property("tankLvl")
        
    @property
    def rechrgM(self):
        return self.get_property("rechrgM")
        
    def calc_mssM(self):
        start_time = self.get_property("mssnStrtTm")
        if start_time:
            return int((datetime.now() - datetime.fromtimestamp(start_time)).total_seconds()//60)
        return None
        
    @property
    def mssnM(self):
        mssM = self.get_property("mssnM")
        if not mssM:
            run_time = self.calc_mssM()
            return run_time if run_time else mssM
        return mssM
    
    @property
    def expireM(self):
        return self.get_property("expireM")
    
    @property
    def cap(self):
        return self.get_property("cap")
    
    @property
    def sku(self) -> str:
        return self.get_property("sku")
        
    @property
    def mission(self):
        return self.get_property("cycle")
        
    @property
    def phase(self):
        return self.get_property("phase")
        
    @property
    def cleanMissionStatus_phase(self):
        return self.phase
        
    @property
    def cleanMissionStatus(self):
        return self.get_property("cleanMissionStatus")
        
    @property
    def pmaps(self):
        return self.get_property("pmaps")
        
    @property
    def regions(self):
        return self.get_property("regions")

    @property
    def map_min_coords(self):
        return self._mapper.min_coords
    
    @property
    def map_max_coords(self):
        return self._mapper.max_coords

    @property
    def docked(self):
        return self.current_state in [
            ROOMBA_STATES["charge"],
            ROOMBA_STATES["recharge"],
            ROOMBA_STATES["dockend"],
            ROOMBA_STATES["evac"]
        ]

    def register_on_message_callback(self, callback):
        self.on_message_callbacks.append(callback)

    def register_on_disconnect_callback(self, callback):
        self.on_disconnect_callbacks.append(callback)

    def _init_remote_client_callbacks(self):
        self.remote_client.set_on_message(self.on_message)
        self.remote_client.set_on_connect(self.on_connect)
        self.remote_client.set_on_disconnect(self.on_disconnect)

    def connect(self):
        if self.roomba_connected or self.periodic_connection_running:
            return

        if self.continuous:
            self._connect()
        else:
            self._thread.daemon = True
            self._thread.start()

        self.time = time.time()  # save connection time

    def _connect(self):
        is_connected = self.remote_client.connect()
        if not is_connected:
            raise RoombaConnectionError(
                "Unable to connect to Roomba at {}".format(
                    self.remote_client.address
                )
            )
        return is_connected

    def disconnect(self):
        if self.continuous:
            self.remote_client.disconnect()
        else:
            self.stop_connection = True

    def periodic_connection(self):
        # only one connection thread at a time!
        if self.periodic_connection_running:
            return
        self.periodic_connection_running = True
        while not self.stop_connection:
            try:
                self._connect()
            except RoombaConnectionError as error:
                self.periodic_connection_running = False
                self.on_disconnect(error)
                return
            time.sleep(self.delay)

        self.remote_client.disconnect()
        self.periodic_connection_running = False

    def on_connect(self, error):
        self.log.info("Connecting to Roomba %s", self.remote_client.address)
        self.client_error = error
        if error is not None:
            self.log.error(
                "Roomba %s connection error, code %s",
                self.remote_client.address,
                error,
            )
            return

        self.roomba_connected = True
        self.remote_client.subscribe(self.topic)

    def on_disconnect(self, error):
        self.roomba_connected = False
        self.client_error = error
        if error is not None:
            self.log.warning(
                "Unexpectedly disconnected from Roomba %s, code %s",
                self.remote_client.address,
                error,
            )

            # call the callback functions
            for callback in self.on_disconnect_callbacks:
                callback(error)

            return

        self.log.info("Disconnected from Roomba %s", self.remote_client.address)

    def on_message(self, mosq, obj, msg):
        if self.exclude != "":
            if self.exclude in msg.topic:
                return

        if self.indent == 0:
            self.master_indent = max(self.master_indent, len(msg.topic))

        log_string, json_data = self._decode_payload(msg.topic, msg.payload)
        self.dict_merge(self.master_state, json_data)

        self.log.debug(
            "Received Roomba Data %s: %s, %s",
            self.remote_client.address,
            str(msg.topic),
            str(msg.payload),
        )

        #update the state machine and history
        #don't update maps if it's not just a signal update
        state = json_data.get("state", {}).get("reported", {})
        should_update_map = len(state) > 1 or "signal" not in state
        
        self._update_state_machine(should_update_map)

        # call the callback functions
        for callback in self.on_message_callbacks:
            callback(json_data)

    def send_command(self, command, params=None):
        if params is None:
            params = {}

        self.log.debug("Send command: %s", command)
        roomba_command = {
            "command": command,
            "time": int(datetime.timestamp(datetime.now())),
            "initiator": "localApp",
        }
        roomba_command.update(params)

        str_command = json.dumps(roomba_command)
        self.log.debug("Publishing Roomba Command : %s", str_command)
        self.remote_client.publish("cmd", str_command)

    def set_preference(self, preference, setting):
        self.log.debug("Set preference: %s, %s", preference, setting)
        val = setting
        # Parse boolean string
        if isinstance(setting, str):
            if setting.lower() == "true":
                val = True
            elif setting.lower() == "false":
                val = False
        tmp = {preference: val}
        roomba_command = {"state": tmp}
        str_command = json.dumps(roomba_command)
        self.log.debug("Publishing Roomba Setting : %s" % str_command)
        self.remote_client.publish("delta", str_command)

    def add_map_definition(self, map: RoombaMap):
        """Adds a map definition"""
        if not map.id:
            raise ValueError(map.id)   
        self._maps[map.id] = map

    def add_map_device(self, device: RoombaMapDevice):
        """Adds a map device"""
        if not device.blid:
            raise ValueError(device.blid)   
        self._devices[device.blid] = device

    def add_map_icon_set(
        self,
        name: str,
        icon_path: str = "{PKG}/assets",                    
        home_icon = None,
        roomba_icon = None,
        error_icon = None,
        cancelled_icon = None,
        battery_low_icon = None,
        charging_icon = None,
        bin_full_icon = None,
        tank_low_icon = None,
        icon_size = DEFAULT_ICON_SIZE,
        show_direction = True):
        """Adds a set of icons for map drawing use"""
        self._mapper.add_icon_set(
            name, icon_path, home_icon, roomba_icon, error_icon,
            cancelled_icon, battery_low_icon,charging_icon,bin_full_icon,tank_low_icon,
            icon_size,show_direction)  
    
    def get_map(self, width: int = None, height: int = None):
        return self._mapper.get_map(width,height)

    def dict_merge(self, dct, merge_dct):
        """
        Recursive dict merge.

        Inspired by :meth:``dict.update()``, instead
        of updating only top-level keys, dict_merge recurses down into dicts
        nested to an arbitrary depth, updating keys. The ``merge_dct`` is
        merged into ``dct``.
        :param dct: dict onto which the merge is executed
        :param merge_dct: dct merged into dct
        :return: None
        """
        for k, v in merge_dct.items():
            if (
                k in dct
                and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], Mapping)
            ):
                self.dict_merge(dct[k], merge_dct[k])
            else:
                dct[k] = merge_dct[k]

    def recursive_lookup(self, search_dict, key, cap=False):
        '''
        recursive dictionary lookup
        if cap is true, return key if it's in the 'cap' dictionary,
        else return the actual key value
        '''
        for k, v in search_dict.items():
            if cap:
                if k == 'cap':
                    return self.recursive_lookup(v, key, False)
            elif k == key:
                return v 
            elif isinstance(v, dict) and k != 'cap':
                val = self.recursive_lookup(v, key, cap)
                if val is not None:
                    return val
        return None

    def get_property(self, property, cap=False):
        '''
        Only works correctly if property is a unique key
        '''
        if property in ['cleanSchedule', 'langs']:
            value = self.recursive_lookup(self.master_state, property+'2', cap)
            if value is not None:
                return value
        return self.recursive_lookup(self.master_state, property, cap)

    def set_flags(self, flags=None):
        self._handle_flags(flags, True)
        
    def clear_flags(self, flags=None):
        self._handle_flags(flags)
        
    def flag_set(self, flag) -> bool:
        try:
            return self._flags.get(flag, False)
        except KeyError:
            pass
        return False
            
    def _handle_flags(self, flags=None, set=False):
        if isinstance(flags, str):
            flags = [flags]
        if flags:
            for flag in flags:
                if set:
                    self._flags[flag] = True
                else:
                    self._flags.pop(flag, None)
        else:
            self._flags = {}

    def _update_history(self, property, value=None, cap=False):
        '''
        keep previous value
        '''
        if value is not None:
            current = value
        else:
            current = self.get_property(property, cap)
        if isinstance(current, dict):
            current = current.copy()
        previous = self._history.get(property, {}).get('current')
        if previous is None:
            previous = current
        self._history[property] = {'current' : current,
                                  'previous': previous}
        return current
        
    def _set_history(self, property, value=None):
        if isinstance(value, dict):
            value = value.copy()
        self._history[property] = {'current' : value,
                                  'previous': value}
        
    def current(self, property):
        return self._history.get(property, {}).get('current')
        
    def previous(self, property):
        return self._history.get(property, {}).get('previous')
        
    def changed(self, property):
        changed = self._history.get(property, {}).get('current') != self._history.get(property, {}).get('previous')
        return changed

    def zero_coords(self, theta=180):
        '''
        returns dictionary with default zero coords
        '''
        return {"x": 0, "y": 0, "theta": theta}
        
    def zero_pose(self, theta=180):
        '''
        returns dictionary with default zero coords
        '''
        return {"theta":theta,"point":{"x":0,"y":0}}  

    def _get_error_message(self, error_num):
        try:
            error_message = ROOMBA_ERROR_MESSAGES[error_num]
        except KeyError as e:
            self.log.warning(
                "Error looking up error message {}".format(e))
            error_message = "Unknown Error number: {}".format(error_num)
        return error_message   

    def _get_not_ready_message(self, not_ready_num):
        try:
            message = ROOMBA_READY_MESSAGES[not_ready_num]
        except KeyError as e:
            self.log.warning(
                "Error looking up not ready message {}".format(e))
            message = "Unknown not ready number: {}".format(not_ready_num)
        return message  

    def _get_map_device(self) -> RoombaMapDevice:
        return self._devices.get(self.blid,None)      

    def _get_mission_map(self) -> RoombaMap:
        return self._get_map(self._pmap_id)

    def _get_map(self, map_id: str) -> RoombaMap:
        try:
            return self._maps[map_id]
        except:
            return self._get_default_map()

    def _get_default_map(self) -> RoombaMap:
        return RoombaMap("default","Default")                     

    def _decode_payload(self, topic, payload):
        """
        Format json for pretty printing.

        Returns string sutiable for logging, and a dict of the json data
        """
        indent = self.master_indent + 31  # number of spaces to indent json data

        json_data = None
        try:
            # if it's json data, decode it (use OrderedDict to preserve keys
            # order), else return as is...
            json_data = json.loads(
                payload.decode("utf-8")
                .replace(":nan", ":NaN")
                .replace(":inf", ":Infinity")
                .replace(":-inf", ":-Infinity"),
                object_pairs_hook=OrderedDict,
            )
            # if it's not a dictionary, probably just a number
            if not isinstance(json_data, dict):
                return json_data, dict(json_data)
            json_data_string = "\n".join(
                (indent * " ") + i
                for i in (json.dumps(json_data, indent=2)).splitlines()
            )

            formatted_data = "Decoded JSON: \n%s" % json_data_string

        except ValueError:
            formatted_data = payload
        return formatted_data, dict(json_data)

    def _update_state_machine(self, should_update_map: bool):
        '''
        Roomba progresses through states (phases), current identified states
        are:
        ""              : program started up, no state yet
        "run"           : running on a Cleaning Mission
        "hmUsrDock"     : returning to Dock
        "hmMidMsn"      : need to recharge
        "hmPostMsn"     : mission completed
        "charge"        : charging
        "stuck"         : Roomba is stuck
        "stop"          : Stopped
        "pause"         : paused
        "evac"          : emptying bin
        "chargingerror" : charging base is unplugged

        available states:
        states = {"charge"          : "Charging",
                  "new"             : "New Mission",
                  "run"             : "Running",
                  "resume"          : "Running",
                  "hmMidMsn"        : "Docking",
                  "recharge"        : "Recharging",
                  "stuck"           : "Stuck",
                  "hmUsrDock"       : "User Docking",
                  "completed"       : "Mission Completed",
                  "cancelled"       : "Cancelled",
                  "stop"            : "Stopped",
                  "pause"           : "Paused",
                  "evac"            : "Emptying",
                  "hmPostMsn"       : "Docking - End Mission",
                  "chargingerror"   : "Base Unplugged",
                  ""                :  None}

        Normal Sequence is "" -> charge -> run -> hmPostMsn -> charge
        Mid mission recharge is "" -> charge -> run -> hmMidMsn -> charge
                                   -> run -> hmPostMsn -> charge
        Stuck is "" -> charge -> run -> hmPostMsn -> stuck
                    -> run/charge/stop/hmUsrDock -> charge
        Start program during run is "" -> run -> hmPostMsn -> charge
        Note: Braava M6 goes run -> hmPostMsn -> run -> charge when docking
        Note: S9+ goes run -> hmPostMsn -> charge -> run -> charge on a training mission (ie cleanMissionStatus_cycle = 'train')
        Note: The first 3 "pose" (x, y) co-ordinate in the first 10 seconds during undocking at mission start seem to be wrong
              for example, during undocking:
              {"x": 0, "y": 0},
              {"x": -49, "y": 0},
              {"x": -47, "y": 0},
              {"x": -75, "y": -11}... then suddenly becomes normal co-ordinates
              {"x": -22, "y": 131}
              {"x": -91, "y": 211}
              also during "hmPostMsn","hmMidMsn", "hmUsrDock" the co-ordinates system also seems to change to bogus values
              For example, in "run" phase, co-ordinates are reported as:
              {"x": -324, "y": 824},
              {"x": -324, "y": 826} ... etc, then some time after hmMidMsn (for example) they change to:
              {"x": 417, "y": -787}, which continues for a while
              {"x": 498, "y": -679}, and then suddenly changes back to normal co-ordinates
              {"x": -348, "y": 787},
              {"x": -161, "y": 181},
              {"x": 0, "y": 0}
              
        Need to identify a new mission to initialize map, and end of mission to
        finalise map.
        mission goes from 'none' to 'clean' (or another mission name) at start of mission (init map)
        mission goes from 'clean' (or other mission) to 'none' at end of missions (finalize map)
        Anything else = continue with existing map
        '''

        mission = self._update_history("cycle")      #mission
        phase = self._update_history("phase")        #mission phase
        self._update_history("pose")                 #update co-ordinates
        self._update_flags()
        self._update_map_id()

        if phase is None or mission is None:
            return
        
        current_mission = self.current_state
        
        self.log.info('current_state: {}, \
            current phase: {}, \
            mission: {}, \
            mission_min: {}, \
            recharge_min: {}, \
            co-ords changed: {}'.format(self.current_state,
                                            phase,
                                            mission,
                                            self.mssnM,
                                            self.rechrgM,
                                            self.changed('pose')))

        try:
            if (self.mssnM is None
                and self.phase == "charge"
                and (
                    self.current_state == ROOMBA_STATES["pause"]
                    or self.current_state == ROOMBA_STATES["recharge"]
                )
            ):
                self.current_state = ROOMBA_STATES["cancelled"]
        except KeyError:
            pass

        if (
            self.current_state == ROOMBA_STATES["charge"]
            and self.phase == "run"
        ):
            self.current_state = ROOMBA_STATES["new"]
            self._new_mission_start_time = time.time()
            self._mapper.reset_map(self._get_mission_map(), self._get_map_device())
        elif (
            self.current_state == ROOMBA_STATES["run"]
            and self.phase == "hmMidMsn"
        ):
            self.current_state = ROOMBA_STATES["dock"]
        elif (
            self.current_state == ROOMBA_STATES["dock"]
            and self.phase == "charge"
        ):
            self.current_state = ROOMBA_STATES["recharge"]
        elif (
            self.current_state == ROOMBA_STATES["recharge"]
            and self.phase == "charge"
            and self.bin_full
        ):
            self.current_state = ROOMBA_STATES["pause"]
        elif (
            self.current_state == ROOMBA_STATES["run"]
            and self.phase == "charge"
        ):
            self.current_state = ROOMBA_STATES["recharge"]
        elif (
            self.current_state == ROOMBA_STATES["recharge"]
            and self.phase == "run"
        ):
            self.current_state = ROOMBA_STATES["pause"]
        elif (
            self.current_state == ROOMBA_STATES["pause"]
            and self.phase == "charge"
            and self.batPct != "100"            
        ):
            self.current_state = ROOMBA_STATES["pause"]
            # so that we will draw map and can update recharge time
            current_mission = None
        elif (
            self.current_state == ROOMBA_STATES["charge"]
            and self.phase == "charge"
            and self.batPct != "100"
        ):
            # so that we will draw map and can update charge status
            current_mission = None
        elif (
            self.current_state == ROOMBA_STATES["stop"]
            or self.current_state == ROOMBA_STATES["pause"]
        ) and self.phase == "hmUsrDock":
            self.current_state = ROOMBA_STATES["cancelled"]
        elif (
            self.current_state == ROOMBA_STATES["hmUsrDock"]
            or self.current_state == ROOMBA_STATES["cancelled"]
        ) and self.phase == "charge":
            self.current_state = ROOMBA_STATES["dockend"]
        elif (
            self.current_state == ROOMBA_STATES["hmPostMsn"]
            and self.phase == "charge"
        ):
            self.current_state = ROOMBA_STATES["dockend"]
        elif (
            self.current_state == ROOMBA_STATES["dockend"]
            and self.phase == "charge"
        ):
            self.current_state = ROOMBA_STATES["charge"]
        else:
            if self.phase not in ROOMBA_STATES:
                self.log.error(
                    "Can't find state %s in predefined Roomba states, "
                    "please create a new issue: "
                    "https://github.com/pschmitt/roombapy/issues/new",
                    self.phase,
                )
                self.current_state = None
            else:
                self.current_state = ROOMBA_STATES[
                    self.phase
                ]

        if self.current_state != current_mission:
            self.log.debug("State updated to: %s", self.current_state)

        #draw the map, forcing a redraw if needed
        if self.current_state and should_update_map:
            self._mapper.update_map(current_mission != self.current_state)

    def _update_flags(self):
        if not self.bin_full:
            self.clear_flags('bin_full')
            
        if self.tanklvl is not None:
            if self.tanklvl < 100:
                self.set_flags('tank_low')
            else:
                self.clear_flags('tank_low')

        if  self.current_state == ROOMBA_STATES["charge"]:
            self.clear_flags(['battery_low', 'stuck'])

        elif self.current_state == ROOMBA_STATES["recharge"]:
            self.clear_flags(['battery_low', 'stuck'])

        elif self.current_state == ROOMBA_STATES["run"]:
            self.clear_flags(['stuck', 'new_mission'])

        elif self.current_state == ROOMBA_STATES["new"]:
            self.clear_flags()
            self.set_flags('new_mission')

        elif self.current_state == ROOMBA_STATES["stuck"]:
            self.set_flags('stuck')

        elif self.current_state == ROOMBA_STATES["cancelled"]:
            self.set_flags('cancelled')

        elif self.current_state == ROOMBA_STATES["hmMidMsn"]:
            if self.bin_full:
                self.set_flags('bin_full')
            else:
                self.set_flags('battery_low')

    def _update_map_id(self):
        try:
            self._pmap_id = self.get_property("pmap_id")
        except:
            pass




DEFAULT_MAP_SKIP_POINTS = 3
DEFAULT_MAP_MAX_ALLOWED_DISTANCE = 500
DEFAULT_BG_COLOR = (0,0,0,0)
DEFAULT_PATH_COLOR = (0,0,180,127)
DEFAULT_TEXT_COLOR = (255,255,255,255)
DEFAULT_TEXT_BG_COLOR = (0,0,0,180)
DEFAULT_IMG_HEIGHT = 1000
DEFAULT_IMG_WIDTH = 1000
DEFAULT_PATH_WIDTH = 2
DEFAULT_MAP_MIN_COORDS = (-1000,-1000)
DEFAULT_MAP_MAX_COORDS = (1000,1000)
DEFAULT_MAP_ANGLE = 0.0

DEFAULT_ICON_PATH = "{PKG}/assets"
DEFAULT_ICON_SIZE = (50,50)
DEFAULT_ICON_HOME = "home.png"
DEFAULT_ICON_ROOMBA = "r865_icon.png"
DEFAULT_ICON_ERROR = "overlay-error.png"
DEFAULT_ICON_CANCELLED = "overlay-cancelled.png"
DEFAULT_ICON_CHARGING = "overlay-charging.png"
DEFAULT_ICON_BATTERY = "overlay-battery-low.png"
DEFAULT_ICON_BIN_FULL = "overlay-bin-full.png"
DEFAULT_ICON_TANK_LOW = "overlay-tank-low.png"



from roombapy import Roomba
from roombapy.remote_client import RoombaRemoteClient


class RoombaFactory:
    """
    Allows you to create Roomba class to control your robot
    """

    @staticmethod
    def create_roomba(
        address=None, blid=None, password=None, continuous=True, delay=1
    ):
        remote_client = RoombaFactory._create_remote_client(
            address, blid, password
        )
        return Roomba(remote_client, continuous, delay)

    @staticmethod
    def _create_remote_client(address=None, blid=None, password=None):
        return RoombaRemoteClient(address=address, blid=blid, password=password)



"""The class helps you to get password for your Roomba."""
import logging
import socket
import ssl
import struct


class RoombaPassword:
    """Main class to get a password."""

    roomba_ip = None
    roomba_port = 8883
    message = None
    server_socket = None
    log = None

    def __init__(self, roomba_ip):
        """Init default values."""
        self.roomba_ip = roomba_ip
        self.message = bytes.fromhex("f005efcc3b2900")
        self.server_socket = _get_socket()
        self.log = logging.getLogger(__name__)

    """
    Roomba have to be on Home Base powered on.
    Press and hold HOME button until you hear series of tones.
    Release button, Wi-Fi LED should be flashing
    After that execute get_password method
    """

    def get_password(self):
        """Get password for roomba."""
        self._connect()
        self._send_message()
        response = self._get_response()
        return _decode_password(response)

    def _connect(self):
        self.server_socket.connect((self.roomba_ip, self.roomba_port))
        self.log.debug(
            "Connected to Roomba %s:%s", self.roomba_ip, self.roomba_port
        )

    def _send_message(self):
        self.server_socket.send(self.message)
        self.log.debug("Message sent")

    def _get_response(self):
        try:
            raw_data = b""
            response_length = 35
            while True:
                if len(raw_data) >= response_length + 2:
                    break

                response = self.server_socket.recv(1024)

                if len(response) == 0:
                    break

                raw_data += response
                if len(raw_data) >= 2:
                    response_length = struct.unpack("B", raw_data[1:2])[0]
            self.server_socket.close()
            return raw_data
        except socket.timeout:
            self.log.warning("Socket timeout")
            return None
        except socket.error as e:
            self.log.debug("Socket error", e)
            return None


def _decode_password(data):
    return str(data[7:].decode().rstrip("\x00"))


def _get_socket():
    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.settimeout(10)
    ssl_socket = ssl.wrap_socket(
        server_socket,
        ssl_version=ssl.PROTOCOL_TLS,
        ciphers="DEFAULT@SECLEVEL=1",
    )
    return ssl_socket



from typing import Tuple, Optional

from .const import (
    DEFAULT_BG_COLOR,
    DEFAULT_PATH_COLOR,
    DEFAULT_PATH_WIDTH
)
from .image_helpers import validate_color

class RoombaMapDevice:
    def __init__(
        self,
        blid: str,
        icon_set: str,
        bg_color,
        path_color,
        path_width
    ):
        self.blid = blid
        self.icon_set = icon_set
        self.bg_color = bg_color
        self.path_color = path_color
        self.path_width = path_width

    @property
    def bg_color(self) -> Tuple[int,int,int,int]:
        return self._bg_color
    
    @bg_color.setter
    def bg_color(self, value):
        self._bg_color = validate_color(value, self._bg_color or DEFAULT_BG_COLOR)
    
    @property
    def path_color(self) -> Tuple[int,int,int,int]:
        return self._path_color
    
    @path_color.setter
    def path_color(self, value):        
        self._path_color = validate_color(value, self._path_color or DEFAULT_PATH_COLOR)

    @property
    def path_width(self) -> int:
        return self._path_width
    
    @path_width.setter
    def path_width(self, value: Optional[int]) -> int:
        if value == None:
            return
        if value <= 0:
            value = DEFAULT_PATH_WIDTH
        self._path_width = value   


import json
import logging
import socket

from roombapy.roomba_info import RoombaInfo


class RoombaDiscovery:
    udp_bind_address = ""
    udp_address = "<broadcast>"
    udp_port = 5678
    roomba_message = "irobotmcs"
    amount_of_broadcasted_messages = 5
    server_socket = None
    log = None

    def __init__(self):
        """Init discovery."""
        self.server_socket = _get_socket()
        self.log = logging.getLogger(__name__)

    def find(self, ip=None):
        if ip is not None:
            return self.get(ip)
        return self.get_all()

    def get_all(self):
        self._start_server()
        self._broadcast_message(self.amount_of_broadcasted_messages)
        robots = set()
        while True:
            response = self._get_response()
            if response:
                robots.add(response)
            else:
                break
        return robots

    def get(self, ip):
        self._start_server()
        self._send_message(ip)
        return self._get_response(ip)

    def _get_response(self, ip=None):
        try:
            while True:
                raw_response, addr = self.server_socket.recvfrom(1024)
                if ip is not None and addr[0] != ip:
                    continue
                self.log.debug(
                    "Received response: %s, address: %s", raw_response, addr
                )
                data = raw_response.decode()
                if self._is_from_irobot(data):
                    return _decode_data(data)
        except socket.timeout:
            self.log.info("Socket timeout")
            return None

    def _is_from_irobot(self, data):
        if data == self.roomba_message:
            return False

        json_response = json.loads(data)
        if (
            "Roomba" in json_response["hostname"]
            or "iRobot" in json_response["hostname"]
        ):
            return True

        return False

    def _broadcast_message(self, amount):
        for i in range(amount):
            self.server_socket.sendto(
                self.roomba_message.encode(), (self.udp_address, self.udp_port)
            )
            self.log.debug("Broadcast message sent: " + str(i))

    def _send_message(self, udp_address):
        self.server_socket.sendto(
            self.roomba_message.encode(), (udp_address, self.udp_port)
        )
        self.log.debug("Message sent")

    def _start_server(self):
        self.server_socket.bind((self.udp_bind_address, self.udp_port))
        self.log.debug("Socket server started, port %s", self.udp_port)


def _decode_data(data):
    json_response = json.loads(data)
    return RoombaInfo(
        hostname=json_response["hostname"],
        robot_name=json_response["robotname"],
        ip=json_response["ip"],
        mac=json_response["mac"],
        firmware=json_response["sw"],
        sku=json_response["sku"],
        capabilities=json_response["cap"],
    )


def _get_socket():
    server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    server_socket.settimeout(5)
    return server_socket



# https://homesupport.irobot.com/app/answers/detail/a_id/21127/kw/charging%20error
MQTT_ERROR_MESSAGES = {
    0: None,
    1: "Bad protocol",
    2: "Bad client id",
    3: "Server unavailable",
    4: "Bad username or password",
    5: "Not authorised",
}

ROOMBA_ERROR_MESSAGES = {
    0: "None",
    1: "Left wheel off floor",
    2: "Main brushes stuck",
    3: "Right wheel off floor",
    4: "Left wheel stuck",
    5: "Right wheel stuck",
    6: "Stuck near a cliff",
    7: "Left wheel error",
    8: "Bin error",
    9: "Bumper stuck",
    10: "Right wheel error",
    11: "Bin error",
    12: "Cliff sensor issue",
    13: "Both wheels off floor",
    14: "Bin missing",
    15: "Reboot required",
    16: "Bumped unexpectedly",
    17: "Path blocked",
    18: "Docking issue",
    19: "Undocking issue",
    20: "Docking issue",
    21: "Navigation problem",
    22: "Navigation problem",
    23: "Battery issue",
    24: "Navigation problem",
    25: "Reboot required",
    26: "Vacuum problem",
    27: "Vacuum problem",
    29: "Software update needed",
    30: "Vacuum problem",
    31: "Reboot required",
    32: "Smart map problem",
    33: "Path blocked",
    34: "Reboot required",
    35: "Unrecognised cleaning pad",
    36: "Bin full",
    37: "Tank needed refilling",
    38: "Vacuum problem",
    39: "Reboot required",
    40: "Navigation problem",
    41: "Timed out",
    42: "Localization problem",
    43: "Navigation problem",
    44: "Pump issue",
    45: "Lid open",
    46: "Low battery",
    47: "Reboot required",
    48: "Path blocked",
    52: "Pad required attention",
    53: "Software update required",
    65: "Hardware problem detected",
    66: "Low memory",
    68: "Hardware problem detected",
    73: "Pad type changed",
    74: "Max area reached",
    75: "Navigation problem",
    76: "Hardware problem detected",
    88: "Back-up refused",
    89: "Mission runtime too long",
    101: "Battery isn't connected",
    102: "Charging error",
    103: "Charging error",
    104: "No charge current",
    105: "Charging current too low",
    106: "Battery too warm",
    107: "Battery temperature incorrect",
    108: "Battery communication failure",
    109: "Battery error",
    110: "Battery cell imbalance",
    111: "Battery communication failure",
    112: "Invalid charging load",
    114: "Internal battery failure",
    115: "Cell failure during charging",
    116: "Charging error of Home Base",
    118: "Battery communication failure",
    119: "Charging timeout",
    120: "Battery not initialized",
    122: "Charging system error",
    123: "Battery not initialized",
}

ROOMBA_READY_MESSAGES = {
    0: 'N/A',
    2: 'Uneven Ground',
    15: 'Low Battery',
    39: 'Pending',
    48: 'Path Blocked'   
}

ROOMBA_STATES = {
    "charge": "Charging",
    "new": "New Mission",
    "run": "Running",
    "resume": "Running",
    "hmMidMsn": "Recharging",
    "recharge": "Recharging",
    "stuck": "Stuck",
    "hmUsrDock": "User Docking",
    "dock": "Docking",
    "dockend": "Docking - End Mission",
    "cancelled": "Cancelled",
    "completed": "Mission Completed",    
    "stop": "Stopped",
    "pause": "Paused",
    "hmPostMsn": "End Mission",
    "evac": "Emptying Bin",
    "chargingerror": "Base Unplugged",
    "": None,
}



import logging
import ssl

import paho.mqtt.client as mqtt

from roombapy.const import MQTT_ERROR_MESSAGES

MAX_CONNECTION_RETRIES = 3


class RoombaRemoteClient:
    address = None
    port = None
    blid = None
    password = None
    log = None
    was_connected = False
    on_connect = None
    on_disconnect = None

    def __init__(self, address, blid, password, port=8883):
        """Create mqtt client."""
        self.address = address
        self.blid = blid
        self.password = password
        self.port = port
        self.log = logging.getLogger(__name__)
        self.mqtt_client = self._get_mqtt_client()

    def set_on_message(self, on_message):
        self.mqtt_client.on_message = on_message

    def set_on_connect(self, on_connect):
        self.on_connect = on_connect

    def set_on_publish(self, on_publish):
        self.mqtt_client.on_publish = on_publish

    def set_on_subscribe(self, on_subscribe):
        self.mqtt_client.on_subscribe = on_subscribe

    def set_on_disconnect(self, on_disconnect):
        self.on_disconnect = on_disconnect

    def connect(self):
        attempt = 1
        while attempt <= MAX_CONNECTION_RETRIES:
            self.log.info(
                "Connecting to %s, attempt %s of %s",
                self.address,
                attempt,
                MAX_CONNECTION_RETRIES,
            )
            try:
                self._open_mqtt_connection()
                return True
            except Exception as e:
                self.log.error(
                    "Can't connect to %s, error: %s", self.address, e
                )
            attempt += 1

        self.log.error("Unable to connect to %s", self.address)
        return False

    def disconnect(self):
        self.mqtt_client.disconnect()

    def subscribe(self, topic):
        self.mqtt_client.subscribe(topic)

    def publish(self, topic, payload):
        self.mqtt_client.publish(topic, payload)

    def _open_mqtt_connection(self):
        if not self.was_connected:
            self.mqtt_client.connect(self.address, self.port)
            self.was_connected = True
        else:
            self.mqtt_client.loop_stop()
            self.mqtt_client.reconnect()
        self.mqtt_client.loop_start()

    def _get_mqtt_client(self):
        mqtt_client = mqtt.Client(client_id=self.blid)
        mqtt_client.username_pw_set(username=self.blid, password=self.password)
        mqtt_client.on_connect = self._internal_on_connect
        mqtt_client.on_disconnect = self._internal_on_disconnect

        self.log.debug("Setting TLS certificate")
        try:
            mqtt_client.tls_set(
                cert_reqs=ssl.CERT_NONE,
                tls_version=ssl.PROTOCOL_TLS,
                ciphers="DEFAULT:!DH",
            )
        except ValueError:  # try V1.3 version
            self.log.warning("TLS Setting failed - trying 1.3 version")
            mqtt_client._ssl_context = None
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
            ssl_context.verify_mode = ssl.CERT_NONE
            ssl_context.set_ciphers("DEFAULT:!DH")
            ssl_context.load_default_certs()
            mqtt_client.tls_set_context(ssl_context)
        mqtt_client.tls_insecure_set(True)

        return mqtt_client

    def _internal_on_connect(self, client, userdata, flags, rc):
        self.log.debug(
            "Connected to Roomba %s, response code = %s", self.address, rc
        )
        connection_error = MQTT_ERROR_MESSAGES[rc]
        if self.on_connect is not None:
            self.on_connect(connection_error)

    def _internal_on_disconnect(self, client, userdata, rc):
        self.log.debug(
            "Disconnected from Roomba %s, response code = %s", self.address, rc
        )
        connection_error = MQTT_ERROR_MESSAGES[rc]
        if self.on_disconnect is not None:
            self.on_disconnect(connection_error)



import sys

from roombapy import RoombaFactory
from roombapy.discovery import RoombaDiscovery
from roombapy.getpassword import RoombaPassword


def discovery():
    roomba_ip = _get_ip_from_arg()

    roomba_discovery = RoombaDiscovery()
    if roomba_ip is not None:
        print(roomba_discovery.find(roomba_ip))
        return

    robots_info = roomba_discovery.find()
    for robot in robots_info:
        print(robot)


def password():
    roomba_ip = _get_ip_from_arg()
    _validate_ip(roomba_ip)
    _wait_for_input()

    roomba_discovery = RoombaDiscovery()
    roomba_info = roomba_discovery.find(roomba_ip)
    _validate_roomba_info(roomba_info)

    roomba_password = RoombaPassword(roomba_ip)
    found_password = roomba_password.get_password()
    roomba_info.password = found_password
    print(roomba_info)


def connect():
    roomba_ip = _get_ip_from_arg()
    _validate_ip(roomba_ip)

    roomba_password = _get_password_from_arg()
    _validate_password(roomba_password)

    roomba_discovery = RoombaDiscovery()
    roomba_info = roomba_discovery.find(roomba_ip)
    _validate_roomba_info(roomba_info)

    roomba = RoombaFactory.create_roomba(
        roomba_info.ip, roomba_info.blid, roomba_password
    )
    roomba.register_on_message_callback(lambda msg: print(msg))
    roomba.connect()

    while True:
        pass


def _validate_ip(ip):
    if ip is None:
        raise Exception("ip cannot be null")


def _validate_password(ip):
    if ip is None:
        raise Exception("password cannot be null")


def _validate_roomba_info(roomba_info):
    if roomba_info is None:
        raise Exception("cannot find roomba")


def _wait_for_input():
    print(
        "Roomba have to be on Home Base powered on.\n"
        "Press and hold HOME button until you hear series of tones.\n"
        "Release button, Wi-Fi LED should be flashing"
    )
    input("Press Enter to continue...")


def _get_ip_from_arg():
    if len(sys.argv) < 2:
        return None
    return str(sys.argv[1])


def _get_password_from_arg():
    if len(sys.argv) < 3:
        return None
    return str(sys.argv[2])



import math
from typing import Tuple

def clamp(num, min_value, max_value):
   return max(min(num, max_value), min_value)

def interpolate(value, in_range, out_range) -> float:
    
    #handle inverted ranges
    invert = False
    if in_range[0] > in_range[1]:
        in_range = in_range[1], in_range[0]
        invert = not invert
    if out_range[0] > out_range[1]:
        in_range = out_range[1], out_range[0]
        invert = not invert

    #make sure it's in the range
    value = clamp(value, in_range[0], in_range[1])

    out = float(value - in_range[0]) / float(in_range[1] - in_range[0]) * float(out_range[1] - out_range[0])
    if invert:
        out = float(out_range[1]) - out

    return out

def rotate(x, y, angle, invert_x: bool = False, invert_y: bool = False) -> Tuple[float,float]:
    xx = x*math.cos(math.radians(angle)) - \
         y*math.sin(math.radians(angle))
    yy = x*math.sin(math.radians(angle)) + \
         y*math.cos(math.radians(angle))    
    
    if invert_x:
        xx = x - (xx - x)
    if invert_y:
        yy = y - (yy - y)
    return xx, yy



import os

def get_mapper_asset(path: str, resource: str):
    if path is None or path.isspace() or path == "":
        return resource
    if path.startswith("{PKG}"):
        return os.path.normpath(os.path.join(os.path.dirname(__file__), path.replace("{PKG}",".."), resource))
    else:
        return os.path.normpath(os.path.join(path, resource))



import logging
from typing import Tuple, Optional

try:
    from PIL import Image
    HAVE_PIL = True
except ImportError:
    pass 

from .const import (
    DEFAULT_BG_COLOR,
    DEFAULT_IMG_HEIGHT,
    DEFAULT_IMG_WIDTH,
    DEFAULT_MAP_ANGLE,
    DEFAULT_MAP_MAX_COORDS,
    DEFAULT_MAP_MIN_COORDS,
    DEFAULT_PATH_COLOR,
    DEFAULT_PATH_WIDTH,
    DEFAULT_TEXT_BG_COLOR,
    DEFAULT_TEXT_COLOR
)
from .image_helpers import make_blank_image, validate_color

class RoombaMap:
    _id: str
    _name: str
    _coords_start: Tuple[int,int] = DEFAULT_MAP_MIN_COORDS
    _coords_end: Tuple[int,int] = DEFAULT_MAP_MAX_COORDS
    _angle: float = DEFAULT_MAP_ANGLE
    _floorplan: Image.Image = None
    _walls: Image.Image = None
    icon_set: str = None
    device: str = None
    
    def __init__(
        self, 
        id,
        name,
        coords_start: Tuple[int,int] = DEFAULT_MAP_MIN_COORDS,
        coords_end: Tuple[int,int] = DEFAULT_MAP_MAX_COORDS,
        angle: float = DEFAULT_MAP_ANGLE,
        floorplan: str = None,
        walls: str = None,
        icon_set: str = None,
        bg_color = DEFAULT_BG_COLOR,
        path_color = DEFAULT_PATH_COLOR,
        path_width = DEFAULT_PATH_WIDTH,
        text_color = DEFAULT_TEXT_COLOR,
        text_bg_color = DEFAULT_TEXT_BG_COLOR):

        self._log = logging.getLogger(__name__)
        self._id = id
        self._name = name
        self.coords_start = coords_start
        self.coords_end = coords_end
        self.angle = angle
        self.icon_set = icon_set
        self.floorplan = floorplan
        self.walls = walls
        self.bg_color = bg_color
        self.path_color = path_color
        self.path_width = path_width
        self.text_color = text_color
        self.text_bg_color = text_bg_color

    @property
    def id(self) -> str:
        return self._id  

    @property
    def name(self) -> str:
        return self._name

    @property
    def coords_start(self) -> Tuple[int,int]:
        return self._coords_start
    
    @coords_start.setter
    def coords_start(self, value):
        self._coords_start = self._validate_coords(value, self._coords_start or DEFAULT_MAP_MIN_COORDS)

    @property
    def coords_end(self) -> Tuple[int,int]:
        return self._coords_start
    
    @coords_end.setter
    def coords_end(self, value):
        self._coords_end = self._validate_coords(value, self._coords_end or DEFAULT_MAP_MAX_COORDS)

    @property
    def angle(self) -> float:
        return self._angle
    
    @angle.setter
    def angle(self, value):
        self._angle = self._validate_angle(value, self._angle or DEFAULT_MAP_ANGLE)        

    @property
    def floorplan(self) -> Image.Image:
        return self._floorplan
    
    @floorplan.setter
    def floorplan(self, value):
        self._floorplan = self._set_image(value)
        if not self._floorplan:
            self._floorplan = make_blank_image(DEFAULT_IMG_WIDTH, DEFAULT_IMG_HEIGHT)

    @property
    def walls(self) -> Image.Image:
        return self._walls
    
    @walls.setter
    def walls(self, value):
        self._walls = self._set_image(value)        

    @property
    def bg_color(self) -> Tuple[int,int,int,int]:
        return self._bg_color
    
    @bg_color.setter
    def bg_color(self, value):
        self._bg_color = validate_color(value, self._bg_color or DEFAULT_BG_COLOR)
    
    @property
    def path_color(self) -> Tuple[int,int,int,int]:
        return self._path_color
    
    @path_color.setter
    def path_color(self, value):        
        self._path_color = validate_color(value, self._path_color or DEFAULT_PATH_COLOR)

    @property
    def path_width(self) -> int:
        return self._path_width
    
    @path_width.setter
    def path_width(self, value: Optional[int]) -> int:
        if value == None:
            return
        self._path_width = value

    @property
    def text_color(self) -> Tuple[int,int,int,int]:
        return self._text_color
    
    @text_color.setter
    def text_color(self, value):
        self._text_color = validate_color(value, self._text_color or DEFAULT_TEXT_COLOR)

    @property
    def text_bg_color(self) -> Tuple[int,int,int,int]:
        return self._text_bg_color      
    
    @text_bg_color.setter
    def text_bg_color(self, value):
        self._text_bg_color = validate_color(value, self._text_bg_color or DEFAULT_TEXT_BG_COLOR)

    @property
    def img_width(self) -> int:
        if self.floorplan:
            x, _ = self.floorplan.size
            return x
        else:
            return DEFAULT_IMG_WIDTH
    
    @property
    def img_height(self) -> int:
        if self.floorplan:
            _, y = self.floorplan.size
            return y
        else:
            return DEFAULT_IMG_HEIGHT

    def _set_image(self, value):
        if value is None:
            return None   
        if isinstance(value, str):
            return Image.open(value).convert('RGBA')
        elif isinstance(value, Image.Image):
            return value
        else:
            self._log.warning(f'Could not load image from {value}')
            return None      

    def _validate_coords(self, value, default) -> Tuple[int,int]:
        if value is None:
            return default
        if not isinstance(value, tuple):
            return default
        if len(value) != 2:
            return default
        return value

    def _validate_angle(self, value, default) -> float:
        if value is None:
            return default
        try:
            v = float(value)
            v %= 360
            if v < 0:
                v += 360
            return v
        except:
            return default


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['roombapy', 'roombapy.mapping']

package_data = \
{'': ['*'], 'roombapy': ['assets/*']}

install_requires = \
['paho-mqtt>=1.5.1,<2.0.0', 'pillow>=8.3.0']

entry_points = \
{'console_scripts': ['roomba-connect = roombapy.entry_points:connect',
                     'roomba-discovery = roombapy.entry_points:discovery',
                     'roomba-password = roombapy.entry_points:password']}

setup_kwargs = {
    'name': 'roombasdk',
    'version': '1.7.7',
    'description': 'Python program and library to control Wi-Fi enabled iRobot Roombas',
    'long_description': '# roombasdk\n\n[![CI](https://github.com/pschmitt/roombasdk/actions/workflows/ci.yaml/badge.svg)](https://github.com/pschmitt/roombasdk/actions/workflows/ci.yaml)\n![PyPI](https://img.shields.io/pypi/v/roombasdk)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/roombasdk)\n![PyPI - License](https://img.shields.io/pypi/l/roombasdk)\n\nUnofficial iRobot Roomba python library (SDK).\n\nFork of [NickWaterton/Roomba980-Python](https://github.com/NickWaterton/Roomba980-Python)<br/>\nFork of [pschmitt/roombapy](https://github.com/pschmitt/roombapy)\n\nThis library was created for the [Home Assistant Roomba integration](https://www.home-assistant.io/integrations/roomba/).\n\n# Installation\n\n```shell\npip install roombasdk\n```\n\n# Notes\n\nThis library is only for firmware 2.x.x [Check your robot version!](http://homesupport.irobot.com/app/answers/detail/a_id/529) \n\nOnly local connections are supported.\n\n# How to get your username/blid and password\n\nTo get password from Roomba type in console:\n\n```shell\n$ roomba-password <ip>\n```\n\nIt will find your Roomba in local network, then follow the instructions in console to get password.\nIf IP address not provided password will be request for auto discovered robot. \n\nAlso you can just ask Roomba for info:\n\n```shell\n$ roomba-discovery <optional ip address>\n```\n\nTo test connection with iRobot:\n\n```shell\n$ roomba-connect <ip> <password>\n```\n\n# Mapping Information\n\nThe Roomba position is given as three coordinates: `x`, `y`, and `theta`.  The unit of measure for `x` and `y` is *cm*, theta is *degrees*.  The origin of the mapping coordinates is the position of the dock, which will have coordinates `(0,0,0)`\n\n## Coordinates \n- Dock Front = -y\n- Dock Back = +y\n- Dock Left = -x\n- Dock Right = -y\n\n```\n         | -y \n         |\n-x -------------- +x\n         |\n         | +y\n```\n\n### Coordinates for Map Definitions\n\nWhen defining maps, you will need to define two points, the upper left `p1` and lower right `p2`.  These coordinates would yield the maximum range for the roomba and will be translated into the image coordinate system automatically.\n\n```\np1       | -y \n         |\n-x -------------- +x\n         |\n         | +y   p2\n```\n\n## Degrees\n\nRoomba reports positive degrees when turning left, and negative degrees when turning right, yielding a counter-clockwise direction.\n\n```\n         0\n         | \n         |\n90 -------------- -90\n         |\n         |\n      -180/180    \n```\n\n',
    'author': 'Philipp Schmitt',
    'author_email': 'philipp@schmitt.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/simbaja/roombapy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Import to public all required modules."""
from .discovery import RoombaDiscovery
from .getpassword import RoombaPassword
from .roomba import Roomba, RoombaConnectionError
from .roomba_factory import RoombaFactory
from .roomba_info import RoombaInfo



import io
import math
import logging
import os
import time
from typing import TYPE_CHECKING, NamedTuple, Tuple
import textwrap

# Import trickery
global HAVE_PIL
HAVE_PIL = False

try:
    from PIL import Image, ImageDraw, ImageFont
    HAVE_PIL = True
except ImportError:
    print("PIL module not found, maps are disabled")

if TYPE_CHECKING:
    from ..roomba import Roomba

from ..const import ROOMBA_STATES
from .const import (
    DEFAULT_BG_COLOR,
    DEFAULT_ICON_SIZE,
    DEFAULT_MAP_MAX_ALLOWED_DISTANCE,
    DEFAULT_MAP_SKIP_POINTS,
    DEFAULT_PATH_COLOR,
    DEFAULT_PATH_WIDTH
)
from .math_helpers import clamp, rotate, interpolate
from .image_helpers import transparent, make_blank_image, center_image
from .misc_helpers import get_mapper_asset
from .roomba_icon_set import RoombaIconSet
from .roomba_map_device import RoombaMapDevice
from .roomba_map import RoombaMap

class RoombaPosition(NamedTuple):
    x: int
    y: int
    theta: int

class MapRenderParameters(NamedTuple):
    icon_set: str
    device: str
    bg_color: Tuple[int,int,int,int]
    path_color: Tuple[int,int,int,int]
    path_width: int
    
class RoombaMapper:
    def __init__(self, 
        roomba: 'Roomba', 
        font: ImageFont.ImageFont = None,

        assets_path = "{PKG}/assets"
    ):
        self.log = logging.getLogger(__name__)
        self.roomba = roomba
        self.map_enabled = False
        self.assets_path = assets_path

        #initialize the font
        self.font = font
        if self.font is None:
            try:
                self.font = ImageFont.truetype(get_mapper_asset(assets_path, "monaco.ttf"), 30)
            except Exception as e:
                self.log.warning(f"Error loading font, loading default font")
                self.font = ImageFont.load_default()

        #generate the default icons
        self._icons: dict[str,RoombaIconSet] = {}
        self.add_icon_set("default"),
        self.add_icon_set("m", roomba_icon="m6_icon.png")
        self.add_icon_set("j", roomba_icon="j7_icon.png")
        self.add_icon_set("s", roomba_icon="s9_icon.png")

        #mapping variables
        self._map: RoombaMap = None
        self._device: RoombaMapDevice = None
        self._render_params: MapRenderParameters = None
        self._rendered_map: Image.Image = None
        self._base_rendered_map: Image.Image = None
        self._points_to_skip = DEFAULT_MAP_SKIP_POINTS
        self._points_skipped = 0
        self._max_distance = DEFAULT_MAP_MAX_ALLOWED_DISTANCE
        self._history = []
        self._history_translated: list[RoombaPosition] = []

        #initialize a base map
        self._initialize_map()

    @property
    def roomba_image_pos(self) -> RoombaPosition:       
        try:
            #roomba sometimes doesn't show the right coords when docked,
            #override the coordinates just in case
            if self.roomba.docked:
                return self._map_coord_to_image_coord(self.roomba.zero_coords())
            return self._history_translated[-1]
        except:
            return RoombaPosition(None,None,None)
    
    @property
    def origin_image_pos(self) -> RoombaPosition:
        return self._map_coord_to_image_coord(self.roomba.zero_coords())

    @property
    def min_coords(self) -> Tuple[int,int]:

        if len(self._history) > 0:
            return (
                min(list(map(lambda p: p["x"], self._history))), 
                min(list(map(lambda p: p["y"], self._history)))
            )
        else:
            return (0,0)

    @property
    def max_coords(self) -> Tuple[int,int]:
        if len(self._history) > 0:
            return (
                max(list(map(lambda p: p["x"], self._history))), 
                max(list(map(lambda p: p["y"], self._history)))
            )
        else:
            return (0,0)          

    def add_icon_set(self,
        name: str,
        icon_path: str = "{PKG}/assets",                    
        home_icon = None,
        roomba_icon = None,
        error_icon = None,
        cancelled_icon = None,
        battery_low_icon = None,
        charging_icon = None,
        bin_full_icon = None,
        tank_low_icon = None,
        icon_size = DEFAULT_ICON_SIZE,
        show_direction = True
    ):
        if not name:
            self.log.error("Icon sets must have names")
            return

        i = RoombaIconSet(size=icon_size, show_direction=show_direction, log=self.log)

        if roomba_icon:
            i.roomba = self._get_mapper_asset(icon_path, roomba_icon)
        if error_icon:
            i.error = self._get_mapper_asset(icon_path, error_icon)
        if cancelled_icon:
            i.cancelled = self._get_mapper_asset(icon_path, cancelled_icon)
        if battery_low_icon:
            i.battery_low = self._get_mapper_asset(icon_path, battery_low_icon)
        if charging_icon:
            i.charging = self._get_mapper_asset(icon_path, charging_icon)
        if bin_full_icon:
            i.bin_full = self._get_mapper_asset(icon_path, bin_full_icon)
        if tank_low_icon:
            i.tank_low = self._get_mapper_asset(icon_path, tank_low_icon)
        if home_icon:
            i.home = self._get_mapper_asset(icon_path, home_icon)

        self._icons[name] = i

    def add_map_device(self, name: str, device: RoombaMapDevice):
        if not name:
            self.log.error("Devices must have names")
            return

        self._devices[name] = device

    def _get_mapper_asset(self, icon_path: str, icon):
        if isinstance(icon, str):
            return get_mapper_asset(icon_path, icon)
        if isinstance(icon, Image.Image):
            return icon
        else:
            return None

    def reset_map(self, map: RoombaMap, device: RoombaMapDevice = None, points_to_skip: int = DEFAULT_MAP_SKIP_POINTS):
        self.map_enabled = self.roomba.cap.get("pose", False) and HAVE_PIL        
        self._history = []
        self._history_translated = []
        self._map = map
        self._device = device
        self._points_to_skip = points_to_skip
        self._points_skipped = 0

        self._initialize_map()

    def _initialize_map(self):
        self._render_params = self._get_render_parameters()

        #generate the base on which other layers will be composed
        base = self._map_blank_image(color=self._render_params.bg_color)

        #add the floorplan if available
        if self._map and self._map.floorplan:
            base = Image.alpha_composite(base, self._map.floorplan)

        #set our internal variables so that we can get the default
        self._base_rendered_map = base
        self._rendered_map = base
    
    def update_map(self, force_redraw = False):
        """Updates the cleaning map"""

        #if mapping not enabled, nothing to update
        if not self.map_enabled:
            return

        if (self.roomba.changed('pose') or self.roomba.changed('phase') or 
            force_redraw) and self.map_enabled:

            #make sure we have phase info before trying to render
            if self.roomba.current_state is not None:
                self._update_state()
                self._render_map()

    def get_map(self, width: int = None, height: int = None) -> bytes:

        #get the default map
        map = self._rendered_map

        #if we haven't rendered anything, just return
        if map is None:
            return None

        #if we have a requested size, resize it
        if width and height:
            map = map.resize((width,height))
            pass 

        #save the internal image
        with io.BytesIO() as stream:
            map.save(stream, format="PNG")
            return stream.getvalue()

    def _update_state(self):
        position: dict[str,int] = None

        if self.roomba.changed('pose'):
            position = self.roomba.co_ords
        
        self.log.debug(f"MAP [State Update]: co-ords: {self.roomba.co_ords} \
                        phase: {self.roomba.phase}, \
                        state: {self.roomba.current_state}")

        if self.roomba.current_state == ROOMBA_STATES["charge"]:
            position = None
        elif self.roomba.current_state == ROOMBA_STATES["evac"]:
            position = None     
        elif self.roomba.current_state == ROOMBA_STATES["completed"]:
            self.log.info("MAP [State Update]: Mission Complete")  
        elif self.roomba.current_state == ROOMBA_STATES["run"]:            
            if self.roomba.co_ords == self.roomba.zero_coords(theta=0):
                #bogus pose received, can't have 0,0,0 when running, usually happens after recovering from an error condition
                self.log.warning('MAP [State Update]: received 0,0,0 pose when running - ignoring')
                position = None

        #if we have a position update, append to the history if it meets our criteria
        if position:
            #there's a few points at the beginning that are usually erroneous, skip them
            if self._points_skipped < self._points_to_skip:
                self._points_skipped += 1
                return

            #if we have history, we need to check a couple things
            if len(self._history) > 0:
                old = self._history[-1]
                old_x = old["x"]
                old_y = old["y"]
                new_x = position["x"]
                new_y = position["y"]

                #if we didn't actually move from the last recorded position, ignore it
                if (old_x,old_y) == (new_x,new_y):
                    return

                #at times, roomba reports erroneous points, ignore if too large of a gap
                #between measurements
                if self._map_distance((old_x,old_y),(new_x,new_y)) > self._max_distance:
                    return

            self._history.append(position)
            self._history_translated.append(self._map_coord_to_image_coord(position))

    def _map_coord_to_image_coord(self, coord: dict) -> RoombaPosition:
        x: float = float(coord["x"])
        y: float = float(coord["y"])
        theta: float = float(coord["theta"])
        
        #perform rotation: occurs about the map origin, so should
        #undo any rotation that exists
        x, y = rotate(x, y, self._map.angle,
            invert_x = self._map.coords_start[0] > self._map.coords_end[0],
            invert_y = self._map.coords_start[1] < self._map.coords_end[1]
        )

        #interpolate the x,y coordinates to scale to the appropriate output
        img_x = interpolate(
            x, 
            [self._map.coords_start[0], self._map.coords_end[0]],
            [0, self._map.img_width - 1]
        )
        img_y = interpolate(
            y, 
            [self._map.coords_start[1], self._map.coords_end[1]],
            [0, self._map.img_height - 1]
        )

        #make sure we stay within the bounds
        clamp(img_x, 0, self._map.img_width)
        clamp(img_y, 0, self._map.img_height)
        
        #adjust theta
        #from what I can see, it looks like the roomba uses a coordinate system:
        #0 = facing away from the dock, increasing angle counter-clockwise
        #it looks like past 180, the roomba uses negative angles, but still seems
        #to be in the counter-clockwise direction
        #PIL denotes angles in the counterclockwise direction
        #so, to compute the right angle, we need to:
        #1) add map angle
        #2) add 180 degrees (roomba image faces up, but should face away at 0)
        #2) add theta
        #3) mod 360, add 360 if negative
        img_theta = (self._map.angle + theta + 180) % 360
        if img_theta < 0:
            img_theta += 360        
        
        #return the tuple
        return RoombaPosition(int(img_x), int(img_y), int(img_theta))

    def _render_map(self):
        """Renders the map"""

        #draw in the vacuum path
        base = self._draw_vacuum_path(self._base_rendered_map)

        #draw in the map walls (to hide overspray)
        if self._map.walls:
            base = Image.alpha_composite(base, self._map.walls)

        #draw the roomba and any problems
        base = self._draw_roomba(base)

        #finally, draw the text
        #base = self._draw_text(base)

        #save the map
        self._rendered_map = base
        
    def _get_render_parameters(self) -> MapRenderParameters:
        if self._map:
            icon_set = self._map.icon_set
            bg_color = self._map.bg_color
            path_color = self._map.path_color
            path_width = self._map.path_width
        else:
            icon_set = self._get_icon_set()
            bg_color = DEFAULT_BG_COLOR
            path_color = DEFAULT_PATH_COLOR
            path_width = DEFAULT_PATH_WIDTH

        if self._device:
            if self._device.icon_set:
                icon_set = self._device.icon_set
            if self._device.bg_color:
                bg_color = self._device.bg_color
            if self._device.path_color:
                path_color = self._device.path_color
            if self._device.path_width:
                path_width = self._device.path_width

        return MapRenderParameters(
            icon_set,
            self._device.blid,
            bg_color,
            path_color,
            path_width
        )

    def _map_blank_image(self, color=transparent) -> Image.Image:
        return make_blank_image(self._map.img_width,self._map.img_height,color)

    def _draw_vacuum_path(self, base: Image.Image) -> Image.Image:
        if len(self._history_translated) > 1:        
            layer = self._map_blank_image()
            renderer = ImageDraw.Draw(layer)

            renderer.line(
                list(map(lambda p: (p.x,p.y), self._history_translated)),
                fill=self._render_params.path_color,
                width=self._render_params.path_width,
                joint="curve"
            )

            return Image.alpha_composite(base, layer)
        else:
            return base

    def _get_icon_set(self):
        #get the default (should always exist)
        icon_set = self._icons["default"]

        #attempt to get the series specific set
        series = self._icons.get(self.roomba.sku[0],None)            
        if series:
            icon_set = series

        #override with the map set if needed
        if self._render_params and self._render_params.icon_set:
            try:
                icon_set = self._icons[self._render_params.icon_set]
            except:
                self.log.warn(f"Could not load icon set '{self._render_params.icon_set}' for map.")

        return icon_set

    def _draw_roomba(self, base: Image.Image, render_params: MapRenderParameters) -> Image.Image:
        layer = self._map_blank_image()

        #get the image coordinates of the roomba
        x, y, theta = self.roomba_image_pos

        #get the icon set to use
        icon_set = self._get_icon_set(render_params)

        #add in the roomba icon
        if x and y:
            rotated = icon_set.roomba.rotate(theta, expand=True)
            layer.paste(rotated, center_image(x, y, rotated, layer.size))

        #add the dock
        dock = self._map_blank_image()
        dock.paste(
            icon_set.home,
            center_image(
                self.origin_image_pos.x, 
                self.origin_image_pos.y, 
                icon_set.home, layer.size
            )
        )

        layer = Image.alpha_composite(layer, dock)

        #add the problem icon (pick one in a priority order)
        problem_icon = None
        if self.roomba._flags.get('stuck'):
            problem_icon = icon_set.error
        elif self.roomba._flags.get('cancelled'):
            problem_icon = icon_set.cancelled
        elif self.roomba._flags.get('bin_full'):
            problem_icon = icon_set.bin_full
        elif self.roomba._flags.get('battery_low'):
            problem_icon = icon_set.battery_low
        elif self.roomba._flags.get('tank_low'):
            problem_icon = icon_set.tank_low

        if x and y and problem_icon:
            problem = self._map_blank_image()
            problem.paste(
                problem_icon,
                center_image(x, y, problem_icon, layer.size)             
            )
            layer = Image.alpha_composite(layer, problem)

        return Image.alpha_composite(base, layer)

    def _draw_text(self, base: Image.Image) -> Image.Image:
        margin = 10

        layer = self._map_blank_image()
        state, attributes, time = self._get_display_text()
        renderer = ImageDraw.Draw(layer)

        #consider something like pynter, perhaps would look better

        combined_text = state.upper()
        if attributes:
            max_len = (base.size[0]-2*margin)//(self.font.getsize(attributes)[0]//len(attributes))
            attributes = textwrap.fill(attributes, max_len)
            combined_text = combined_text + "\n" + attributes
        if time:
            combined_text = combined_text + "\n" + "Time: " + time            
        
        #get the bounding box
        bbox = renderer.multiline_textbbox((margin,margin), combined_text, self.font)
        bbox = (bbox[0]-margin,bbox[1]-margin,bbox[2]+margin,bbox[3]+margin)

        #render a background box
        renderer.rectangle(bbox, fill=self._map.text_bg_color)

        #render the text
        renderer.multiline_text((margin,margin), combined_text, fill=self._map.text_color, font=self.font)

        return Image.alpha_composite(base, layer)

    def _get_display_text(self) -> Tuple[str,str,str]:
        display_state: str = None
        display_attributes: str = None
        display_time: str = None
        show_time: bool = False

        if  self.roomba.current_state == ROOMBA_STATES["charge"]:
            display_state = "Charging"
            display_attributes = f"Battery: {self.roomba.batPct}%"
        elif self.roomba.current_state == ROOMBA_STATES["recharge"]:
            display_state = "Recharging"
            display_attributes = f"Time: {self.roomba.rechrgM}m, \
                             Bat: {self.roomba.batPct}%"
        elif self.roomba.current_state == ROOMBA_STATES["pause"]:
            display_state = "Paused"
            display_attributes = f"{self.roomba.mssnM}m, \
                             Bat: {self.roomba.batPct}%"
        elif self.roomba.current_state == ROOMBA_STATES["hmPostMsn"]:
            display_state = "Returning Home"
            show_time = True
        elif self.roomba.current_state == ROOMBA_STATES["evac"]:
            display_state = "Emptying Bin"
        elif self.roomba.current_state == ROOMBA_STATES["completed"]:
            display_state = "Completed"
            show_time = True            
        elif self.roomba.current_state == ROOMBA_STATES["run"]:
            display_state = "Running"
            display_attributes = f"Time {self.roomba.mssnM}m, Bat: {self.roomba.batPct}%"
        elif self.roomba.current_state == ROOMBA_STATES["stop"]:
            display_state = "Stopped"
            display_attributes = f"Time {self.roomba.mssnM}m, Bat: {self.roomba.batPct}%"
        elif self.roomba.current_state == ROOMBA_STATES["new"]:
            display_state = "Starting"
        elif self.roomba.current_state == ROOMBA_STATES["stuck"]:
            expire = self.roomba.expireM
            expire_text = 'Job Cancel in {expire}m' if expire else 'Job Cancelled'
            display_state = "Stuck"
            display_attributes = f"{self.roomba.error_message} {expire_text}"
            show_time = True
        elif self.roomba.current_state == ROOMBA_STATES["cancelled"]:
            display_state = "Cancelled"
            show_time = True
        elif self.roomba.current_state == ROOMBA_STATES["hmMidMsn"]:
            display_state = "Docking"
            display_attributes = f"Bat: {self.roomba.batPct}%, Bin Full: {self.roomba.bin_full}"
        elif self.roomba.current_state == ROOMBA_STATES["hmUsrDock"]:
            display_state = "User Docking"
            show_time = True
        else:
            display_state = self.roomba.current_state
        
        if show_time:
            display_time = time.strftime("%a %b %d %H:%M:%S")

        return display_state, display_attributes, display_time    

    def _map_distance(self, pos1: 'Tuple[int,int]', pos2: 'Tuple[int,int]'):
        return int(math.sqrt(((pos2[0]-pos1[0])**2)+((pos2[1]-pos1[1])**2)))  

    def _interpolate_path_color(f_co, t_co, interval):
        det_co =[(t - f) / interval for f , t in zip(f_co, t_co)]
        for i in range(interval):
            yield [round(f + det * i) for f, det in zip(f_co, det_co)]  


import logging
from typing import Tuple

try:
    from PIL import Image, ImageDraw, ImageFont, ImageColor
    HAVE_PIL = True
except ImportError:
    pass 

from .const import (
    DEFAULT_ICON_BATTERY,
    DEFAULT_ICON_BIN_FULL,
    DEFAULT_ICON_CANCELLED,
    DEFAULT_ICON_CHARGING,
    DEFAULT_ICON_ERROR,
    DEFAULT_ICON_HOME,
    DEFAULT_ICON_PATH,
    DEFAULT_ICON_ROOMBA,
    DEFAULT_ICON_SIZE,
    DEFAULT_ICON_TANK_LOW
)
from .misc_helpers import get_mapper_asset

class RoombaIconSet:
    def __init__(self, 
        size: Tuple[int,int] = DEFAULT_ICON_SIZE, 
        show_direction: bool = True,
        log: logging.Logger = None):

        if log:
            self._log = log
        else:
            self._log = logging.getLogger(f"Roomba.{__name__}")
        
        self.size = size
        self.show_direction = show_direction
        self._icons: dict[str,Image.Image] = {}
        self._load_defaults()

    @property
    def roomba(self) -> Image.Image:
        return self._icons["roomba"]
    
    @roomba.setter
    def roomba(self, value):
        self._set_icon("roomba", value)

    @property
    def error(self) -> Image.Image:
        return self._icons["error"]
    
    @error.setter
    def error(self, value):
        self._set_icon("error", value)
    
    @property
    def cancelled(self) -> Image.Image:
        return self._icons["cancelled"]
    
    @cancelled.setter
    def cancelled(self, value):
        self._set_icon("cancelled", value)

    @property
    def battery_low(self) -> Image.Image:
        return self._icons["battery-low"]
    
    @battery_low.setter
    def battery_low(self, value):
        self._set_icon("battery-low", value)

    @property
    def charging(self) -> Image.Image:
        return self._icons["charging"]
    
    @charging.setter
    def charging(self, value):
        self._set_icon("charging", value)

    @property
    def bin_full(self) -> Image.Image:
        return self._icons["bin-full"]
    
    @bin_full.setter
    def bin_full(self, value):
        self._set_icon("bin-full", value)

    @property
    def tank_low(self) -> Image.Image:
        return self._icons["tank-low"]
    
    @tank_low.setter
    def tank_low(self, value):
        self._set_icon("tank-low", value)

    @property
    def home(self) -> Image.Image:
        return self._icons["home"]
    
    @home.setter
    def home(self, value):
        self._set_icon("home", value)

    def _load_defaults(self):
        self._load_icon_file("roomba", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_ROOMBA))
        self._load_icon_file("error", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_ERROR))
        self._load_icon_file("cancelled", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_CANCELLED))
        self._load_icon_file("battery-low", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_BATTERY))
        self._load_icon_file("charging", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_CHARGING))
        self._load_icon_file("bin-full", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_BIN_FULL))
        self._load_icon_file("tank-low", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_TANK_LOW))
        self._load_icon_file("home", get_mapper_asset(DEFAULT_ICON_PATH, DEFAULT_ICON_HOME))

    def _set_icon(self, name, value):
        if value is None:
            return    
        if isinstance(value, str):
            self._load_icon_file(name, value)
            self._draw_direction(name)
        elif isinstance(value, Image.Image):
            resized = value.convert('RGBA').resize(self.size,Image.ANTIALIAS)
            self._icons[name] = resized
            self._draw_direction(name)
        else:
            raise ValueError()

    def _load_icon_file(self, name, filename, size=None):
        try:
            if not size:
                size = self.size
            icon = Image.open(filename).convert('RGBA').resize(
                size,Image.ANTIALIAS)
                    
            self._icons[name] = icon
        except IOError as e:
            self._log.warning(f'Error loading icon file: {filename}: {e}')

    def _draw_direction(self, name):
        icon = self._icons[name]
        if name == "roomba" and self.show_direction:
            draw_icon = ImageDraw.Draw(icon)
            draw_icon.pieslice([(5,5),(icon.size[0]-5,icon.size[1]-5)],
                265, 275, fill="red", outline="red")